from pyEELSMODEL.misc import hs_gdos as hsdos
from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge
from pyEELSMODEL.database.Zhang.download import download_file
import os
import h5py
from pyEELSMODEL import __file__


class ZezhongCoreLossEdge(CoreLossEdge):
    """
    Coreloss edges which are calculated by Zezhong Zhang.
    https://zenodo.org/records/11199911

    """

    def __init__(self, specshape, A, E0, alpha, beta, element, edge, eshift=0,
                 q_steps=100, dir_path=None):
        """

        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        A : float
              Amplitude of the edge

        E0: float [V]
              The acceleration voltage of the microscope

        alpha: float [rad]
              The convergence angle of the incoming probe

        beta: float [rad]
              The collection angle

        element: string
              The name of the element from which to calculate the edge model.

        edge: string
              The type of edge. (K, L or M)

        eshift: float [eV]
              The shift of the onset energy with respect to the literature
              value.
            (default: 0)

        q_steps: int
             The number of q points taken into account for the integration
             over the momentum space. The larger the number of q_steps the more
             accurate the calculation. (default: 100)

        Returns
        -------
        """
        if dir_path is None:
            self.dir_path = os.path.dirname(
                os.path.dirname(__file__) + "/../pyEELSMODEL/database/Zhang/"
            )
            self.file = os.path.join(self.dir_path, 'Dirac_GOS.gosh')    
        else:
            self.set_dir_path(dir_path)
        if not os.path.exists(self.file):
                download_file(filename=self.file)        
        if edge == 'K':
            edge = 'K1'

        super().__init__(specshape, A, E0, alpha, beta, element, edge,
                         eshift=eshift, q_steps=q_steps)
        self.set_gos_energy_q()
        self.swap_axes = False
        # an approximation for the GOS which does not do the full integration
        self.use_approx = False

    def set_gos_energy_q(self):
        with h5py.File(self.file, 'r') as f:
            self.gos = f[self.element][self.edge]['data'][:].squeeze().T
            self.free_energies = f[self.element][self.edge]['free_energies'][:][
                                 :]  # two dimensional array
            self.q_axis = f[self.element][self.edge]['q'][
                          :] # in  [1/m]
            self.ionization_energy = f[self.element][self.edge]['metadata'].attrs['ionization_energy']

    def set_element(self, element):
        self.element = element

        # file = os.path.join(self.dir_path, element + '.hdf5')
        # isExist = os.path.exists(file)
        # if not isExist:
        #     raise ValueError('Element you selected is not valid')
        # self.element = element
        # self.dir_path_element = file  # the filename to have easy access

    def set_dir_path(self, path):
        self.dir_path = path
        self.file = os.path.join(path, 'Dirac_GOS.gosh')

    def _check_str_in_list(self, list, edge):
        for name in list:
            if name[0] == edge[0]:
                return True

        return False

    def set_edge(self, edge):
        """
        Checks if the given edge is valid and adds the directories of the
        :param edge:
        :return:
        """
        edge_list = ['K1', 'L1', 'L2', 'L3', 'M1', 'M2', 'M3', 'M4', 'M5',
                     'N1', 'N2', 'N3', 'N4', 'N5', 'N6', 'N7']
        if not isinstance(edge, str):
            raise TypeError(
                'Edge should be a string: K1, L1, L2, L3, M2, M3, M4, M5, N4, '
                'N5,N6, N7')
        if edge in edge_list:
            self.edge = edge
        else:
            raise ValueError(
                'Edge should be: K1, L1, L2, L3, M2, M3, M4, M5, N4, N5',
                'N6, N7')

    def calculate_cross_section(self):
        ek = self.onset_energy
        E0 = self.parameters[1].getvalue()
        alpha = self.parameters[3].getvalue()
        beta = self.parameters[2].getvalue()
        # e_axis = self.free_energies[:,0]
        e_axis = self.free_energies

        q_axis = self.q_axis
        gos = self.gos

        pref = 1e28 * self.prefactor

        if self.use_approx:
            cross_section = pref * hsdos.dsigma_dE_from_GOSarray_approx(
                self.energy_axis, e_axis + ek, E0, beta, gos)
        else:
            # cs = hsdos.dsigma_dE_from_GOSarray(self.energy_axis, e_axis + ek,
            #                                    ek+e_axis[0], E0, beta, alpha,
            #                                    q_axis,
            #                                    gos, q_steps=self.q_steps,
            #                                    swap_axes=self.swap_axes)

            cs = hsdos.dsigma_dE_from_GOSarray_bound(self.energy_axis, e_axis,
                                                     ek, E0, beta, alpha,
                                                     q_axis, gos, q_steps=100)

            cross_section = pref * cs

        return cross_section



import logging
import numpy as np
import matplotlib.pyplot as plt

from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.operators.zlpremoval import ZLPRemoval
from pyEELSMODEL.core.spectrum import Spectrum
from pyEELSMODEL.core.multispectrum import MultiSpectrum

logger = logging.getLogger(__name__)


class ThicknessEstimator(Operator):
    """
    Calculates the thickness of the material.
    """
    def __init__(self, spectrum, model_type='Lorentzian', signal_range=None):
        self.spectrum = spectrum
        self.model_type = model_type
        self.signal_range = signal_range

    def log_ratio_method(self):
        """
        Calculates the inelastic mean free path by esitmating the zero loss
        peak. The estimated zlp is stored in the .zlp attribute. The inelastic
        mean free path is stored in the .tlambda attribute
        """
        if self.model_type == 'Mirrored':
            zlpremoval = ZLPRemoval(self.spectrum, self.signal_range)
            zlpremoval.mirrored_zlp()
            self.zlpremoval = zlpremoval
            self.zlp = self.zlpremoval.mirror_zlp

        elif self.model_type == 'Vacuum':
            sub = self.spectrum.get_interval(self.signal_range)
            int_vac = self.vacuum.get_interval(self.signal_range).integrate()
            int_sub = sub.integrate()
            fc1 = self.vacuum.data[np.newaxis, np.newaxis, :]
            fc2 = int_sub[:, :, np.newaxis]
            zlp_data = int_vac**(-1) * fc1 * fc2

            self.zlp = MultiSpectrum.from_numpy(zlp_data,
                                                self.spectrum.energy_axis)

        else:
            zlpremoval = ZLPRemoval(self.spectrum, self.signal_range,
                                    self.model_type)
            zlpremoval.fit_zlp()
            self.zlpremoval = zlpremoval
            self.zlp = self.zlpremoval.zlp

        It = self.spectrum.multidata.sum(2)  # total integral
        Izlp = self.zlp.multidata.sum(2)  # zero loss integral
        self.tlambda = np.log(It / Izlp)

    def extract_vacuum(self, threshold, show_inelastic=True):
        """
        It can happen that vacuum is present and this can be used to
        estimate the thickness. The vacuum spectrum is stored as the
        .vacuum attribute

        Parameters
        ----------
        threshold : 0 < float < 1
            The number of times the spectrum needs to be upsampled
        show_inelastic: boolean
            Indicates whether the inelastic signal used for vacuum
            determination is used.
        """
        fwhm = self.spectrum.get_numerical_fwhm()
        print('FWHM is {} eV'.format(fwhm))
        print('Integral for inelastic signal start at {} eV'.format(5*fwhm))
        Ei = 3*fwhm
        Ef = self.spectrum.energy_axis[-1]
        inelastic = self.spectrum.integrate((Ei, Ef))

        co = int(threshold * inelastic.size)
        thres = np.sort(inelastic.flatten())[co]
        boolean = inelastic < thres

        avg = self.spectrum.multidata[boolean].mean(0)
        s = Spectrum.from_numpy(avg, self.spectrum.energy_axis)
        self.vacuum = s

        if show_inelastic:
            fig, ax = plt.subplots(1, 3)
            ax[0].imshow(inelastic)
            ax[1].imshow(boolean)
            ax[2].plot(self.vacuum.energy_axis, self.vacuum.data,
                       label='vacuum')
            ax[2].plot(self.spectrum.energy_axis, self.spectrum.mean().data,
                       label='average')
            ax[2].legend()

    # def zlp_integral(self):
    #     """
    #     Calculates the integral of the ZLP. If an analytical
    #     function is used for the fitting, this expression can
    #     used. Else a numerical summation is used. Note that
    #     a better implementation of the integral can be developed.
    #     """
    #     if type(self.spectrum) is MultiSpectrum:
    #         A = self.zlpremoval.fitter.coeff_matrix[:, :, 0]
    #         width = self.zlpremoval.fitter.coeff_matrix[:, :, 2] #fwhm
    #     elif type(self.spectrum) is Spectrum:
    #         A = self.zlpremoval.fitter.coeff[0]
    #         width = self.zlpremoval.fitter.coeff[2]
    #
    #     if self.model_type == 'Gaussian':
    #         sigma = np.abs(width) / (2.0 * np.sqrt(2.0 * np.log(2.0)))
    #         zlp_integral = np.sqrt(2) * A * sigma
    #
    #     elif self.model_type == 'Lorentzian':
    #         a = 1
    #     elif self.model_type == 'Voigt':
    #         a = 1
    #     elif self.model_type == 'Mirrored':
    #         zlp = self.zlpremoval.mirror_zlp.sum()
    #         zlp_integral = zlp.integrate()



# -*- coding: utf-8 -*-
"""
This script calculates shows as small script on how to use the fixed pattern
and perform a linear least squares fit
@author: joverbee
"""

import sys
from pyEELSMODEL.components.fixedpattern import FixedPattern
from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.core.spectrum import Spectrumshape
from pyEELSMODEL.core.spectrum import Spectrum
from pyEELSMODEL.core.model import Model
from pyEELSMODEL.fitters.linear_fitter import LinearFitter
import numpy as np

sys.path.append("..")  # Adds higher directory to python modules path.


def test_fixed_pattern():
    """
    Test where a linear fit is performed on the sum of gaussian which both
    will be a fixedpattern.
    :return:
    """
    specshape = Spectrumshape(1, 100, 1024)
    ga1 = Gaussian(specshape, 100, 500, 100)
    ga2 = Gaussian(specshape, 250, 800, 50)
    ga1.calculate()
    ga2.calculate()

    ndata = ga1.data + ga2.data + np.random.normal(0, 1, ga1.data.size)
    s = Spectrum(specshape, data=ndata)

    specshape1 = Spectrumshape(0.33, 50, 3*1500)

    ga1 = Gaussian(specshape1, 1, 500, 100)
    ga1.calculate()

    pat1 = Spectrum(specshape1, data=ga1.data)
    comp1 = FixedPattern(specshape, pat1)
    comp1.calculate()

    specshape2 = Spectrumshape(2, 200, int(626/2))

    ga2 = Gaussian(specshape2, 1, 800, 50)
    ga2.calculate()

    pat2 = Spectrum(specshape2, data=ga2.data)
    comp2 = FixedPattern(specshape, pat2)
    comp2.calculate()

    mod = Model(specshape)
    mod.addcomponent(comp1)
    mod.addcomponent(comp2)

    s.set_exclude_region(825, 1023)

    fit = LinearFitter(s, mod)
    fit.perform_fit()
    fit.set_fit_values()

    assert (np.abs(fit.coeff[0]-100)/100) < 1e-2
    assert (np.abs(fit.coeff[1]-250) / 250) < 1e-2


def test_fixed_pattern_gradient():
    specshape = Spectrumshape(1, 100, 1024)
    specshape1 = Spectrumshape(0.33, 50, 3 * 1500)

    ga1 = Gaussian(specshape1, 1, 500, 100)
    ga1.calculate()

    pat1 = Spectrum(specshape1, data=ga1.data)
    comp1 = FixedPattern(specshape, pat1)
    comp1.parameters[0].setvalue(10)
    comp1.calculate()

    grad = comp1.getgradient(comp1.parameters[0])
    assert np.all(np.abs(comp1.data/10 - grad) < 1e-3)


def main():
    test_fixed_pattern()
    test_fixed_pattern_gradient()


if __name__ == "__main__":
    main()



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np
from pyEELSMODEL import __file__
import os
import h5py
import logging
import pyEELSMODEL.misc.hydrogen_gdos as hdos

logger = logging.getLogger(__name__)


class CoreLossEdge(Component):
    """
    CoreLossEdge generic component class to derive core loss edge
    implementations from.


    """

    def __init__(self, specshape, A, E0, alpha, beta, element, edge, eshift=0,
                 q_steps=100):
        """
        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        A : float
            Amplitude of the edge

        E0: float [V]
            The acceleration voltage of the microscope in volts

        alpha: float [rad]
            The convergence angle of the incoming probe in radians,

        beta: float [rad]
            The collection angle in radians.

        element: string
            The name of the element from which to calculate the edge model.

        edge: string
            The type of edge. (K1, L1, L2, L3, M1, etc.)

        eshift: float [eV]
            The shift of the onset energy with respect to the literature value.
            (default: 0)

        q_steps: int
            The number of q points taken into account for the integration over
            the momentum space. The larger the number of q_steps the more
            accurate the calculation. (default: 100)

        """
        super().__init__(specshape)

        self.elements_dir = os.path.dirname(__file__)
        # print(self.elements_dir)
        self.elements_name = 'element_info.hdf5'

        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(-np.inf, np.inf)
        p1.sethasgradient(True)
        self._addparameter(p1)

        p2 = Parameter('E0', E0, changeallowed=False)
        self._addparameter(p2)

        p3 = Parameter('beta', beta, changeallowed=False)
        self._addparameter(p3)

        p4 = Parameter('alpha', alpha, changeallowed=False)
        self.parameters.append(p4)

        self.eshift_ = eshift
        self.set_element(element)
        self.set_Z()
        self.set_edge(edge)
        self.set_onset_energy()
        self.q_steps = q_steps
        name = element + ' ' + edge + ' edge' + str(self.onset_energy) + ' eV'
        self.setdisplayname(name)

    def get_elements_dir(self):
        return os.path.join(self.elements_dir, self.elements_name)

    def get_elements(self):
        """
        Returns a list of allowed element names
        """
        with h5py.File(self.get_elements_dir(), 'r') as f:
            elem_list = list(f.keys())
        return elem_list

    def set_element(self, element):
        """
        Sets given element to the CoreLossEdge.

        Parameters
        ----------
        element: string
            The element to which to set it.

        """
        ky = self.get_elements()
        if element in ky:
            self.element = element
        else:
            raise ValueError('Not a valid element, here are some examples of '
                             'valid elements: {}, {},'
                             ' {}'.format(ky[0], ky[10], ky[30]))

    def set_Z(self):
        """
        Set the atomic number Z of the element from the CoreLossEdge.
        Needed for the hydrogenic calculations
        """
        with h5py.File(self.get_elements_dir(), 'r') as f:
            self.Z = f[self.element].attrs['Z']

    def set_onset_energy(self):
        with h5py.File(self.get_elements_dir(), 'r') as f:
            onset = f[self.element][self.edge].attrs['onset_energy']
            onset += self.eshift_
            prf = f[self.element][self.edge].attrs['occupancy_ratio']

            self.onset_energy = onset
            self.prefactor = prf

    def set_edge(self, edge):
        """
        Checks if the given edge is valid and adds the directories of the
        :param edge:
        :return:
        """
        edge_list = ['K1', 'L1', 'L2', 'L3', 'M1', 'M2', 'M3', 'M4', 'M5',
                     'N1', 'N2', 'N3', 'N4', 'N5', 'N6', 'N7']
        if not isinstance(edge, str):
            raise TypeError(
                'Edge should be a string: K1, L1, L2, L3, M2, M3, M4, M5, N4,'
                ' N5', 'N6', 'N7')
        if edge in edge_list:
            self.edge = edge
        else:
            raise ValueError(
                'Edge should be: K1, L1, L2, L3, M2, M3, M4, M5, N4, N5', 'N6',
                'N7')

    def calculate_cross_section(self):
        print('if you see this you have called calculate_cross section '
              'on a coreloss edge that didnt implement '
              'calculate_cross_section')

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        pA = self.parameters[0]
        p2 = self.parameters[1]
        p3 = self.parameters[2]
        p4 = self.parameters[3]

        if p2.ischanged() or p3.ischanged() or p4.ischanged():
            self.cross_section = self.calculate_cross_section()
            self.data = pA.getvalue() * self.cross_section

        if pA.ischanged():
            self.data = pA.getvalue() * self.cross_section
        self.setunchanged()

    def getgradient(self, parameter):
        # gradient with respect to weight is easy, others are too difficult
        pA = self.parameters[0]
        if parameter == pA:
            self.gradient[0] = self.cross_section
            return self.gradient[0]
        else:
            return None

    def get_convergence_correction_factor(self, nsamples=100):
        """
        Calculates the convergence correction for each different theta values.
        This depends on the geometry of the setup.

        :return:
        """
        alpha = self.parameters[3].getvalue()
        beta = self.parameters[2].getvalue()
        theta_array = np.exp(
            np.linspace(np.log(1e-9), np.log(alpha + beta), nsamples))
        F = hdos.convergence_correction_factor(alpha, beta, theta_array)
        return F, theta_array



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
import numpy as np
from scipy.optimize import least_squares

from pyEELSMODEL.core.fitter import Fitter
import logging
logger = logging.getLogger(__name__)


class LSQFitter(Fitter):
    """
    The  least squares fitter which can be used to solve non-linear
    problems by optimizing the least squares.
    """

    def __init__(self, spectrum, model, use_bounds=False, method='lm'):
        """
        Initialises a Least Squares Fitter instance.

        Parameters
        ----------
        spectrum : Spectrum or Multispectrum
            The experimental data used which needs to be fitted
        model : Model
            The model used to fit the experimental data. Model should only
            contain
            linear parameters.
        use_bounds: bool, optional
            Bool indicating which if the boundaries of the parameters are used
            (default:False)
        method: string
            The optimization methods used, candiates are lm, trf and dogbox.
            See https://docs.scipy.org/doc/scipy/reference/generated/
            scipy.optimize.least_squares.html for more information.

        Returns
        -------
        An instance of a LSQFitter.

        """

        super().__init__(spectrum, model)
        self.estimator = self.residuals
        self.method = method
        self.use_bounds = use_bounds
        self.use_weights = False

    @property
    def estimator(self):
        return self._estimator

    @estimator.setter
    def estimator(self, estimator):
        self._estimator = estimator

    @property
    def use_bounds(self):
        return self._use_bounds

    @use_bounds.setter
    def use_bounds(self, b):
        if self.method == 'lm':
            logger.info(r'lm algorithm cannot use bounds')
            self._use_bounds = False
        else:
            self._use_bounds = b

    def residuals(self, x0, m, s):
        for index, param in enumerate(m.getfreeparameters()):
            param.setvalue(x0[index])
        m.calculate()

        if self.use_weights:
            # weights = 1/(np.sqrt(np.abs(s.data[np.invert(s.exclude)])+0.01))
            # weights = self.weights[np.invert(s.exclude)]
            weights = self.weights[np.invert(s.exclude)]
            # weights = np.sqrt(np.abs(s.data[np.invert(s.exclude)]))
        else:
            weights = np.ones(s.data[np.invert(s.exclude)].size)

        inv_ = np.invert(s.exclude)

        res = np.sqrt(weights)*(s.data[inv_] - m.data[inv_])
        return res

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method):
        # todo select which methods can be used an tell which needs to have
        methods = ['trf', 'lm', 'dogbox']
        if method in methods:
            self._method = method
        else:
            raise ValueError(r'Given method does not exist for this function')

    def bounds_lsq(self):
        """
        Boundaries for lsq is different as used for the ml fitter

        """
        bounds = self.get_bounds()
        low = []
        up = []
        for bound in bounds:
            low.append(bound[0])
            up.append(bound[1])
        return [tuple(low), tuple(up)]

    def perform_fit(self):

        x0 = self.get_start_parameters()

        for index, param in enumerate(self.model.getfreeparameters()):
            param.setvalue(x0[index])
        self.model.calculate()
        ndata = np.copy(self.spectrum.data)
        ndata[ndata < 1] = 1
        self.weights = 1/ndata

        argo = (self.model, self.spectrum)
        if self.use_bounds:
            bounds = self.bounds_lsq()
        else:
            bounds = (-np.inf), np.inf

        resLS = least_squares(self.estimator, x0, args=argo,
                              method=self.method, bounds=bounds, max_nfev=100)

        self.coeff = resLS.x
        self.error = resLS.cost



import os


def test_file_exists():
    filename = os.path.join('pyEELSMODEL', 'element_info.hdf5')
    assert os.path.exists(filename)


def main():
    test_file_exists()


if __name__ == "main":
    main()



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 29 16:25:55 2021

@author: joverbee
"""

import sys
from pyEELSMODEL.components.polynomial import Polynomial
from pyEELSMODEL.core.spectrum import Spectrumshape
import numpy as np
sys.path.append("..")  # Adds higher directory to python modules path.


def test_polynomial():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = Polynomial(specshape, order=3)  # create a component
    values = [-0.1, 20, 5, 3]
    for param, value in zip(bkg.parameters, values):
        param.setvalue(value)
    bkg.calculate()

    E = bkg.energy_axis

    norm3 = (E ** 3).sum()
    norm2 = (E ** 2).sum()
    norm1 = (E ** 1).sum()
    norm0 = (E ** 0).sum()

    poly = -(0.1 / norm3) * E ** 3 + (20 / norm2) * E ** 2 + (
                5 / norm1) * E + 3 / norm0

    assert np.all(np.abs(poly - bkg.data) < 1e-6)


def test_polynomial_gradient():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = Polynomial(specshape, order=3)  # create a component
    values = [-0.1, 20, 5, 3]
    for param, value in zip(bkg.parameters, values):
        param.setvalue(value)
    bkg.calculate()

    E = bkg.energy_axis
    orders = [3, 2, 1, 0]
    for param, order in zip(bkg.parameters, orders):
        grad = bkg.getgradient(param)
        assert np.all(np.abs(E ** order - grad) < 1e-6)


def main():
    test_polynomial()
    test_polynomial_gradient()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  9 20:46:09 2021

@author: joverbee
"""
import logging
logger = logging.getLogger(__name__)


class Operator():
    """
    Operator class from which all other operations on the spectrum class are
    derived
    """
    def __init__(self):
        self.name = 'Operator'



from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.operators.areaselection import AreaSelection
from pyEELSMODEL.operators.quantification.elemental_quantification import \
    ElementalQuantification
from pyEELSMODEL.components.gdoslin import GDOSLin
from pyEELSMODEL.components.fixedpattern import FixedPattern
import matplotlib.pyplot as plt


class ExperimentalEdgeExtractor(Operator):
    """
    Extract experimental edge from an EELS map which can be used to
    do oxidation state mapping or other things.

    """
    def __init__(self, multispectrum, settings, ll):
        """
        Initialize a ExperimentalEdgeExractor object.
        Parameters
        ----------
        multispectrum: MultiSpectrum
            The multispectrum from which the reference edges will be extracted
        settings: tuple
            Tuple containing the acceleration voltage [V], convergence angle
             [rad] and collection angle [rad]
        ll: MultiSpectrum
            The low loss multispectrum which is acquired from the same region.
            This is needed to incorporate the effect of multiple scattering.
        """

        self.multispectrum = multispectrum
        self.ll = ll
        self.areas = []
        self.settings = settings

    def define_new_region(self, input_map=None, max_points=4, coords=None):
        """
        Define an AreaSelection object. This object is able to extract the
        average spectrum from a given area which can be selected via the
        openend image. This is done by right clicking on the image.
        One could also provide coords which then selects the area from
        the given coordinates.

        Parameters
        ----------
        input_map: numpy 2d array
            The image which is shown. Could be the ADF image acquired simul-
            taneously. If None, the image shows the total intensity of the
            detector at each probe position. (default: None)
        max_points: uint
            The number of points used to draw the area.
        coords: list
            A list with x and y-coordinates of the points. For instance for
            a rectangle with starting position (1,2), width=5 and height=7
            coords = [[1,1,6,6],[2,9,9,2]].

        """
        if coords is None:
            area = AreaSelection(self.multispectrum, input_map=input_map,
                                 other_spectra=[self.ll],
                                 max_points=max_points)
            area.determine_input_area()
        else:
            area = AreaSelection(self.multispectrum, input_map=input_map,
                                 other_spectra=[self.ll],
                                 max_points=max_points)
            area.xcoords = coords[0]
            area.ycoords = coords[1]

        self.areas.append(area)

    def show_regions(self):
        """
        Shows the selected areas together with the index. This index is
        usefull for the self.extract_edge() function.

        Returns
        -------
        fig: Figure
            The create figure
        """
        fig, ax = plt.subplots()
        ax.imshow(self.areas[0].input_map, cmap='gray')
        for ii, area in enumerate(self.areas):
            ax.fill(area.xcoords, area.ycoords, alpha=0.8)
            ax.scatter(area.xcoords, area.ycoords, color='black')
            ax.text(area.xcoords[0], area.ycoords[0], 'Area: '+str(ii),
                    fontsize=16, color='red')
        return fig

    def calculate_spectra(self):
        """
        Calculates the average spectra from the areas which have been
        selected. These are stored in the self.spectra and self.llspectra
        attributes.
        """
        self.spectra = []
        self.llspectra = []
        for area in self.areas:
            s = area.get_mean_from_area()
            self.spectra.append(s)
            self.llspectra.append(area.other_avg_spec[0])

    def show_average_spectra(self):
        """
        Shows the average spectra from the different areas.

        Returns
        -------
        fig: Figure
            The create figure
        """

        fig, ax = plt.subplots()
        for ii, spec in enumerate(self.spectra):
            ax.plot(spec.energy_axis, spec.data, label='Area: '+str(ii))

    def add_quantification_method(self,  hl, ll, elements, edges, intervals,
                                  pre_fine):
        """
        Creates a quantification object. This could be used to modify some
        settings to optimize the fit itself.

        Parameters
        ----------
        hl: Spectrum
            The spectrum from which to get the reference edge
        ll: Spectrum
            The low loss spectrum which is from the same part as hl.
        elements: list
            List of elements which are present in the spectrum hl
        edges: list
            List of edge of the elements. The length of this list should be
            equal to the elements list.
        intervals: list
            List of energy intervals [eV] over which to fit the fine structure.
        pre_fine: list
            List of energies onset energy with respect to the onset energy
            of the atomic cross section. Hence a positive value indicates
            that the fine structure starts before expected onset from the
            atomic cross section. Negative values would be unexpected.
        """
        quant = ElementalQuantification(hl, elements, edges, self.settings,
                                        ll=ll)
        quant.linear_fitter_method = 'ols'
        quant.n_bgterms = 4
        quant.pre_fine = pre_fine
        quant.use_fine = True
        quant.fine_intervals = intervals
        quant.do_align = False
        self.quant = quant

    def extract_edge(self, index, elements, edges, intervals, pre_fine):
        """
        Fits the average spectrum indicated by index using the elements and
        edges. The interval indicates the energy region over which to fit
        the fine structure. Pre_fine is used to modify the start position
        of the fine structure with respect to the atomic cross section. The
        sampling of the fine structure is determined by the fwhm of the zlp.

        Parameters
        ----------
        index: int
            The index of the chosen area. The first area selected has index 0
            and so forth. The show_regions() function also shows which area
            corresponds to which index.
        elements: list
            List of elements which are present in the spectrum of index
        edges: list
            List of edge of the elements. The length of this list should be
            equal to the elements list.
        intervals: list
            List of energy intervals [eV] over which to fit the fine structure.
        pre_fine: list
            List of energies onset energy with respect to the onset energy
            of the atomic cross section. Hence a positive value indicates
            that the fine structure starts before expected onset from the
            atomic cross section. Negative values would be unexpected.
        """
        hl = self.spectra[index]
        ll = self.llspectra[index]

        self.add_quantification_method(hl, ll, elements, edges, intervals,
                                       pre_fine)
        self.quant.do_procedure()

        self.quant.fitter.plot()

        fixs = []
        for ii in range(len(elements)):
            fix = self._get_norm_cross_section(ii)
            fixs.append(fix)

        return fixs

    def _get_norm_cross_section(self, index):
        """
        Calulates the normalized cross section from the fit. It puts the
        atomic cross sections amplitude A=1 and scales the fine structure
        with this. The new reference edge is the sum of atomic and fine
        structure and is a FixedPattern object.

        Parameters
        ----------
        index: uint
            The index from the elements in the fitter. This is needed to
            extract the atomic cross section and fine structure for an edge.

        Returns
        -------
        fix: FixedPattern
            The new edge which is the sum of atomic cross section and fine
            structure.

        """

        component = self.quant.element_components[index]

        fines = []
        for comp in self.quant.fitter.model.components:
            check_gdos = isinstance(comp, GDOSLin)

            if check_gdos:
                print(comp.connected_edge)
                check_connected_edge = comp.connected_edge is component
                print(check_connected_edge)
                if check_connected_edge:
                    print('here')
                    fines.append(comp)

        new_comp = component.copy()
        for fine in fines:
            new_comp.data += fine.data

        new_comp.data = new_comp.data / component.parameters[0].getvalue()
        fix = FixedPattern(new_comp.get_spectrumshape(), new_comp,
                           name=component.name)
        fix.calculate()
        return fix



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np

def dec2bin(n, idx):
    if n > 1:
        dec2bin(n // 2, idx)

    idx.append(n % 2)


def harmonic_space(x0, x1, n, eps=1e-3):

    lnx0 = np.log(x0)
    lnx1 = np.log(x1)
    x = np.exp(np.linspace(lnx0, lnx1, n))
    ii_test = n // 2

    for jj in range(100):
        x_test = x[ii_test] + 0.
        for ii in range(1, len(x) - 1):
            x[ii] = 2. * x[ii-1] * x[ii+1] / (x[ii-1] + x[ii+1])

        if np.abs(x[ii_test] - x_test) < eps:
            break

    return x


class LinearBG(Component):
    """
    Fast background model, where the number of components can be varied.
    Convexity constraints are added to the model 

    Parameters
    ----------
    specshape : Spectrumshape
        The spectrum shape used to model
    rlist : list
        List of the r values used in the linear background model.
        (default: [1,2,3,4,5])
    constrains : str
        The inequality constraints used in the background model. 
        One can chose between 'sufficient', 'necessary', 'non-neg'.

    Returns
    -------
    """

    def __init__(self, specshape, rlist=[1, 2, 3, 4, 5],
                 constrains='sufficient'):
        super().__init__(specshape)

        n = len(rlist)
        for i in range(n):
            pname = 'a' + str(i)
            p = Parameter(pname, 1.0, True)
            p.setboundaries(-np.inf, np.inf)
            # is this true as we will multiply this with another cross section
            p.setlinear(True)
            self._addparameter(p)

            qname = 'r' + str(i)
            q = Parameter(qname, rlist[i], changeallowed=False)
            q.sethasgradient(False)
            self._addparameter(q)

        self.n = n  # number of terms in the sum

        # don't convolute the background it only gives problems and
        # adds no extra physics
        self._setcanconvolute(False)
        self._setshifter(False)  # it is not a shifter type of component

        self.npoints = 3
        self.use_approx = constrains #which approximation for convexity to use
        self._setname('Linear constrained background')

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        changes = False
        for param in self.parameters:
            changes = changes or param.ischanged()
        if changes:
            Alist = []
            rlist = []
            for i in range(self.n):
                p = self.parameters[2 * i]
                Alist.append(p.getvalue())
                q = self.parameters[2 * i + 1]
                rlist.append(q.getvalue())

            self.data = self.linear_background(Alist, rlist)
        self.setunchanged()  # put parameters to unchanged

    def linear_background(self, Alist, rlist):
        E = self.energy_axis
        Estart = E[0]
        if Estart < 1:
            Estart = 1

        mask = (E > 0)  # only meaningfull when E>0 and r>0
        signal = np.zeros(E.size)
        for i in range(len(Alist)):
            signal += mask * (Alist[i] * (E / Estart) ** (-rlist[i]))
        return signal

    def getgradient(self, parameter):
        """calculate the analytical partial derivative wrt parameter j
        returns true if succesful, gradient is stored in component.gradient
        """
        # todo implement the analytical gradient
        for ii, param in enumerate(self.parameters):
            if param is parameter:
                index = ii

        if parameter in self.parameters[::2]:
            r = self.parameters[index + 1].getvalue()
            partial = self.linear_background([1], [r])
            return partial

        else:
            return None

    def get_rlist(self):
        rlist = []
        for param in self.parameters:
            if param.linear:
                pass
            else:
                rlist.append(param.getvalue())
        return rlist
    
    def get_convex_boundaries(self):
        if (self.use_approx == None):
            return None

        # 'relaxed' is legacy
        if (self.use_approx.lower() == 'sufficient') \
                or (self.use_approx.lower() == 'relaxed'):
            return self.convexity_constraints_sufficient()
        # 'approx' is legacy
        elif (self.use_approx.lower() == 'necessary') \
                or (self.use_approx.lower() == 'approx'):
            return self.convexity_constraints_necessary()

        elif (self.use_approx.lower() == 'non-neg'):
            return self.convexity_constraints_nonneg()
        
        else:
            print('Error: \'' + self.use_approx + '\' is not a valid mode.')
            print('Use \'sufficient\', \'necessary\', or \'non-neg\'.')
            return None

    def convexity_constraints_sufficient(self):
        '''
        :param x_01: array with start and endpoint of energy range
        :param r: array with the exponents, the model is E^-r. So do not forget
        about the minus sign
        :return: G: matrix with the constraints
        '''

        r = np.asarray(self.get_rlist())
        delta_r = r[0] - r

        x0 = self.energy_axis[0]
        x1 = self.energy_axis[-1]

        r_m = np.abs(delta_r[1:])
        xh = ((r_m + 1.) / r_m) * (x0 ** (-r_m) - x1 ** (-r_m)) \
             / (x0 ** (-r_m - 1.) - x1 ** (-r_m - 1.))
        xh = np.mean(xh)

        n = len(r)
        pf = r * (r + 1.) * (x0**r)  # added the E0 in the formula DJ
        pf = np.reshape(pf, (1, n))

        G = np.zeros((2 ** (n - 1) + 2, n)) + pf

        for ii in range(2 ** (n - 1)):
            idx0 = []
            dec2bin(ii, idx0)
            idx = np.zeros((n - 1,))

            # Fill out the zeros. Start from the back.
            for jj in range(len(idx0)):
                kk = -jj - 1
                idx[kk] = idx0[kk]

            # Fill out G. Start from the back.
            for jj in range(n - 1):
                kk = -jj - 1
                if idx[kk] == 0:
                    if ii == 0:
                        G[ii, kk] *= x0 ** delta_r[kk]
                    else:
                        G[ii, kk] *= xh ** delta_r[kk]
                else:
                    if ii == (2 ** (n - 1) - 1):
                        G[ii, kk] *= x1 ** delta_r[kk]
                    else:
                        G[ii, kk] *= x1 ** delta_r[kk] * \
                                     (1. + (1. - xh / x1) * (r[kk] + 2.))

        # Value at the end of the axis is positive
        G[-1, :] = (x0 ** r) * x1 ** delta_r
        # Derivative at the end of the axis is negative
        G[-2, :] = r * (G[-1, :] + 0.)

        return G

    def convexity_constraints_necessary(self):
        '''
        :param x_01: array with start and endpoint of energy range
        :param r: array with the exponents, the model is E^-r. So do not forget
         about the minus sign
        :param m: number of constraints.
        :param flag_decr: set to 1 if the 1st derivative
        :return: G: matrix with the constraints
        '''

        r = np.asarray(self.get_rlist())
        x0 = self.energy_axis[0]  # can be changed when only fitting a subregion
        x1 = self.energy_axis[-1]

        m = max((self.npoints, 2))  # at least two points for forced convexity
        # in approx method
        x = harmonic_space(x0, x1, m)  # m is guaranteed to be 2 or higher

        n = len(r)
        pf = r * (r + 1.) * (x0**r)  # pre-factor
        delta_r = r[0] - r
        G = np.zeros((m + 2, n)) + np.reshape(pf, (1, n))

        for ii in range(0, m):
            G[ii, 1:] *= x[ii] ** delta_r[1:]
        # Value at the end of the axis is positive
        G[-1, :] = (x0 ** r) * x1 ** delta_r
        # Derivative at the end of the axis is negative
        G[-2, :] = r * (G[-1, :] + 0.)

        return G

    def convexity_constraints_nonneg(self):
        '''
        :param x_01: array with start and endpoint of energy range
        :param r: array with the exponents, the model is E^-r. So do not forget
        about the minus sign
        :param m: number of constraints.
        :param flag_decr: set to 1 if the 1st derivative
        :return: G: matrix with the constraints
        '''

        r = np.asarray(self.get_rlist())
        n = len(r)

        G = np.eye(n)

        return G


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 22 17:58:57 2022

@author: joverbee
"""

from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
   name='pyEELSMODEL',
   version='1.0.4',
   author='Jo Verbeeck and Daen Jannis',
   author_email='jo.verbeeck@uantwerpen.be',
   package_dir={'': '.'},
   packages=find_packages(),
   url='https://github.com/joverbee/pyEELSMODEL',
   license='LICENSE.txt',
   description='a model based quantification library for electron energy loss spectroscopy',
   long_description=long_description,
   long_description_content_type='text/markdown',
   install_requires=[
       "Django >= 1.1.1",
       "h5py",
       "pqdm",
       "numpy>=1.20.0",
       "matplotlib",
       "scipy",
       "tqdm",
       "fpdf",
       "pyqt5",
       "pytest",
       "jupyter",
       "requests",
       "quadprog"
   ],
   setup_requires=['flake8'],
   test_requires=['pytest'],
   package_data={"pyEELSMODEL": ["element_info.hdf5"]},
   # exclude_package_data={"pyEELSMODEL.database.Zhang": ["*.hdf5"]},
   exclude_package_data={'pyEELSMODEL': ['database/Zhang/*.hdf5',]},

)


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 29 16:25:55 2021

@author: joverbee
"""

import sys
from pyEELSMODEL.components.exponential import Exponential
from pyEELSMODEL.core.spectrum import Spectrumshape
import numpy as np
sys.path.append("..")  # Adds higher directory to python modules path.


def test_exponential():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = Exponential(specshape, 1000, -0.001)  # create a component
    bkg.calculate()

    exp = 1000*np.exp(bkg.energy_axis*-0.001)
    assert np.all(np.abs(exp-bkg.data)) < 1e-6


def test_exponential_gradient():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = Exponential(specshape, 1000, -0.001)  # create a component
    bkg.calculate()

    gradA = bkg.getgradient(bkg.parameters[0])
    gradA_exp = np.exp(bkg.energy_axis*-0.001)
    assert np.all(np.abs(gradA-gradA_exp)) < 1e-6

    gradb = bkg.getgradient(bkg.parameters[1])
    gradb_exp = bkg.energy_axis*1000*np.exp(bkg.energy_axis*-0.001)
    assert np.all(np.abs(gradb-gradb_exp)) < 1e-6


def test_exponential_autofit():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = Exponential(specshape, 1000, -0.001)  # create a component
    bkg.calculate()

    bkg1 = Exponential(specshape, 1, -1)  # create a component
    bkg1.autofit(bkg, 0, 500)

    assert np.abs(bkg1.parameters[0].getvalue() - 1000) < 1e-3
    assert np.abs(bkg1.parameters[1].getvalue() + 0.001) < 1e-3


def main():
    test_exponential()
    test_exponential_gradient()
    test_exponential_autofit()


if __name__ == "__main__":
    main()



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
import numpy as np
from pyEELSMODEL.core.parameter import Parameter


class Lorentzian_sq(Component):
    def __init__(self, specshape, A, centre, fwhm):
        super().__init__(specshape)

        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(0, np.inf)
        p1.sethasgradient(False)
        self._addparameter(p1)

        p2 = Parameter('centre', centre)
        p2.setboundaries(-np.inf, np.inf)
        p2.sethasgradient(False)
        self._addparameter(p2)

        p3 = Parameter('FWHM', fwhm)
        p3.setboundaries(0, np.inf)
        p3.sethasgradient(False)
        self._addparameter(p3)

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        E = self.energy_axis
        self.data = self.lorentz_sq(E, self.parameters[0].getvalue(),
                                    self.parameters[1].getvalue(),
                                    self.parameters[2].getvalue())

    def lorentz_sq(self, x, A, x0, dT):
        # return A*(dT**2/4)*(1/((x-x0)**2+(0.5*dT)**2))
        return A * (1 / ((x - x0) ** 2 + (0.5 * dT) ** 2))**2

    def getgradient(self, parameter):
        return None



from pyEELSMODEL.components.gdoslin import GDOSLin
from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge
import pyEELSMODEL.misc.physical_constants as pc
import numpy as np


class ConstrainedGDOSLin(GDOSLin):
    def __init__(self,  specshape, estart, ewidth=50, degree=21,
                 interpolationtype='quadratic', beta=50e-3, alpha=1e-9,
                 E0=300e3, connected_edge=None):

        super().__init__(specshape, estart, ewidth, degree, interpolationtype)

        self.connected_edge = connected_edge
        self.beta = beta
        self.E0 = E0
        self.alpha = alpha

        self.calculate_integral()

    @classmethod
    def gdoslin_from_edge(cls, specshape, component, pre_e=5, ewidth=50,
                          degree=21, interpolationtype='quadratic'):
        """
        Class method is made to create an gdoslin which is connected to a
        coreloss edge. No need to input the onset energy
        """

        if not isinstance(component, CoreLossEdge):
            raise TypeError(r'Component should be a CoreLossEdge')

        estart = component.onset_energy - pre_e

        beta = component.parameters[2].getvalue()
        alpha = component.parameters[3].getvalue()
        E0 = component.parameters[1].getvalue()

        comp = ConstrainedGDOSLin(specshape, estart, ewidth=ewidth,
                                  degree=degree,
                                  interpolationtype=interpolationtype,
                                  beta=beta, alpha=alpha, E0=E0,
                                  connected_edge=component)
        comp.set_gdos_name()
        return comp

    def calculate_prefactor(self):
        """
        This is the prefactor which determines the constrains used for
        our work
        """

        theta_E = pc.characteristic_angle(self.energy_axis, self.E0,
                                          use_rel=True)
        print('collection angle is: ' + str(self.beta))
        prefactor = self.energy_axis \
                    / np.log(1 + self.beta ** 2 / theta_E ** 2)
        prefactor = prefactor / prefactor[0]
        self.prefactor = prefactor #de behoudswet voor onze GOS

    def calculate_convergent_prefactor(self):
        """
        Calculates the prefactor when a convergent probe is used. The integral
        needs to be solved numerically due to the factor coming from the
        convergent probe.
        """

        F, theta = self.connected_edge.get_convergence_correction_factor(nsamples=1000)
        theta_E = pc.characteristic_angle(self.energy_axis, self.E0,
                                          use_rel=True)

        int_angle = np.zeros_like(self.energy_axis)
        for ii in range(self.energy_axis.size):
            dtheta = np.diff(theta)
            int_angle[ii] = np.sum(F[:-1]*dtheta*2*np.pi*theta[:-1]
                                   /(theta[:-1]**2 + theta_E[ii]**2))

        print('convergence angle is: ' + str(self.alpha))
        print('collection angle is: ' + str(self.beta))


        prefactor = self.energy_axis/int_angle
        prefactor = prefactor / prefactor[0]
        self.convergent_prefactor = prefactor


    def calculate_integral(self):
        """
        Test case when bethe sum rule says that  NOT the total cross section
        is conserved but (Ei-En)*sigma(E) integrated is conserved. This
        modifies the constrains in a straightforward why but needs to know the
        integral of each basis function x dE. This is what gets calculated

        :return:
        """
        if self.alpha > 1e-6:
            print('Convergent incoming beam is used')
            self.calculate_convergent_prefactor()
        else:
            print('Parallel incoming beam is used')
            self.calculate_prefactor()

        self.integral = []
        for ii, param in enumerate(self.parameters[2:]):
            for par in self.parameters[2:]:
                if par == param:
                    par.setvalue(1)
                else:
                    par.setvalue(0)
            self.calculate()

            if self.alpha > 1e-6:
                res = self.data*self.convergent_prefactor
            else:
                res = self.data*self.prefactor
            # res = self.data
            self.integral.append(res.sum())

        for param in self.parameters[2:]:
            param.setvalue(1)

        self.integral = np.array(self.integral)[np.newaxis, :]

    def total_atomic_cross_section(self):
        """
        Calculate the total atomic cross section times the prefactor to have
        a fair comparison to evaluate the constrains or too apply some
        inequality constrains.
        """
        index0 = self.get_energy_index(self.parameters[0].getvalue())
        index1 = self.get_energy_index(self.parameters[0].getvalue()
                                       + self.parameters[1].getvalue())

        if self.alpha > 1e-6:
            atomic_sum = np.sum(self.convergent_prefactor[index0:index1] *
                                self.connected_edge.cross_section[
                                index0:index1])
        else:
            atomic_sum = np.sum(self.prefactor[index0:index1] *
                                self.connected_edge.cross_section[
                                index0:index1])


        self.atomic_sum = atomic_sum

    def get_equality_constraints(self):
        '''
        The sum to zero could be interpreted as a equality constrain.
        :return:
        '''
        return self.integral


    def shadow(self, specshape, onset_shift=6, ratio=.5):
        '''
        Returns a dependent copy of itself applied to specshape
        '''

        onset = self.parameters[0].getvalue()
        interval = self.parameters[1].getvalue()

        shadowed = ConstrainedGDOSLin(specshape, estart=onset + onset_shift,
                                      ewidth=interval, degree=self.degree,
                                      interpolationtype=self.interpolationtype)

        for ii in range(self.degree):
            shadowed.parameters[ii + 2].couple(self.parameters[ii + 2],
                                               fraction=ratio)

        return shadowed

    def check_equality_constrain(self):
        """
        Function which checks if the values
        :return:
        """
        values = np.zeros(self.integral.shape[1])
        for ii, param in enumerate(self.parameters[2:]):
            values[ii] = param.getvalue()

        sum_rule = np.sum(values * self.integral[0])

        print('The equality constrain should put the value to '
              'zero: {}'.format(sum_rule))


import requests
import os 

url = 'https://zenodo.org/records/7645765/files/Segger_Guzzinati_Kohl_1.5.0.gosh'
dir_path = os.path.dirname(os.path.dirname(__file__) + "/../pyEELSMODEL/database/Segger_Guzzinati_Kohl/")
filename = dir_path + 'Segger_Guzzinati_Kohl_1.5.0.gosh'

def download_file(url = url, filename = filename):
    # Send a GET request to the URL
    response = requests.get(url, stream=True)
    
    # Check if the request was successful
    if response.status_code == 200:
        # Open a local file with write-binary mode
        with open(filename, 'wb') as f:
            # Write the content to the file in chunks
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
        print(f"Downloaded '{filename}' successfully.")
    else:
        print(f"Failed to download file. Status code: {response.status_code}")


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 29 16:25:55 2021

@author: joverbee
"""

import sys

from pyEELSMODEL.components.powerlaw import PowerLaw
from pyEELSMODEL.core.spectrum import Spectrumshape
import numpy as np
sys.path.append("..")  # Adds higher directory to python modules path.


def test_powerlaw():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = PowerLaw(specshape, 1000, 3)  # create a component
    bkg.calculate()

    pw = 1000 * (bkg.energy_axis / bkg.energy_axis[0]) ** (-3)
    assert np.all(np.abs(pw - bkg.data)) < 1e-6


def test_powerlaw_gradient():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = PowerLaw(specshape, 1000, 3)  # create a component
    bkg.calculate()

    gradA = bkg.getgradient(bkg.parameters[0])
    gradA_exp = (bkg.energy_axis / bkg.energy_axis[0]) ** (-3)
    assert np.all(np.abs(gradA - gradA_exp)) < 1e-6

    gradb = bkg.getgradient(bkg.parameters[1])
    gradb_exp = 1000 * (bkg.energy_axis / bkg.energy_axis[0]) ** (-3) * (
        -1) * np.log(bkg.energy_axis / bkg.energy_axis[0])
    assert np.all(np.abs(gradb - gradb_exp)) < 1e-6


def test_powerlaw_autofit():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = PowerLaw(specshape, 1000, 3)  # create a component
    bkg.calculate()

    bkg1 = PowerLaw(specshape, 1, 1)  # create a component
    bkg1.autofit(bkg, 0, 500)

    assert np.abs(bkg1.parameters[0].getvalue() - 1000) < 1e-3
    assert np.abs(bkg1.parameters[1].getvalue() - 3) < 1e-3


def main():
    test_powerlaw()
    test_powerlaw_gradient()
    test_powerlaw_autofit()


if __name__ == "__main__":
    main()



import numpy as np
from pyEELSMODEL.misc import physical_constants as pc

global XU, IE3, IE1, EK
XU = (0.52, 0.42, 0.30, 0.29, 0.22, 0.30, 0.22, 0.16, 0.12, 0.13, 0.13, 0.14,
      0.16, 0.18, 0.19, 0.22, 0.14, 0.11, 0.12, 0.12, 0.12, 0.1, 0.1, 0.1)
IE3 = (73.0, 99.0, 135.0, 164.0, 200.0, 245.0, 294.0, 347.0, 402.0, 455.0,
       513.0, 575.0, 641.0, 710.0, 779.0, 855.0, 931.0, 1021.0, 1115.0,
       1217.0, 1323.0, 1436.0, 1550.0, 1675.0)


IE1 = (118.0, 149.0, 189.0, 229.0, 270.0, 320.0, 377.0, 438.0, 500.0, 564.0,
       628.0, 695.0, 769.0, 846.0, 926.0, 1008.0, 1096.0, 1194.0, 1142.0,
       1248.0, 1359.0, 1476.0, 1596.0, 1727.0)
EK = (13, 25, 55, 111, 188, 284, 402, 532, 685, 867, 1072, 1305, 1560, 1839,
      2146, 2472, 2822, 3203, 3607, 4038, 4493, 4966)


def get_XU():
    return XU


def get_IE3():
    return IE3


def get_IE1():
    return IE1


def get_EK():
    return EK


def gdos_k(E, qa0sq, Z):
    """
    Calculates the hydrogenic GDOS for the k edge of an atom with atomic number
    Z.

    Parameters
    ----------
    E : float
        The energy in eV.
    qa0sq : float
        The q*a0**2 value (derived from momentum)
    Z : int
        Atomic number
    Returns
    -------
        GDOS
    """
    Zs = pc.Zs_k(Z)
    Q = (qa0sq / Zs ** 2)
    kHsq = E / (Zs ** 2 * pc.R()) - 1
    akH = np.sqrt(np.abs(kHsq))
    eps = 1e-99  # to not devide by zero
    # implemented in hyperspy (maybe value to small gives erros)
    if akH < 0.01:
        akH = 0.01

    if kHsq >= 0.0:

        beta = np.arctan(2 * akH / (Q - kHsq + 1))
        if beta < 0:
            beta = beta + np.pi
        teller = 256 * E * (Q + kHsq / 3 + 1 / 3) * np.exp(-2 * beta / akH)
        noemer = Zs ** 4 * pc.R() ** 2 * (
                    (Q - kHsq + 1) ** 2 + (4 * kHsq)) ** 3 * (
                             1 - np.exp(-2 * np.pi / akH)) + eps
        df_dE = teller / noemer
    else:
        y = -(1 / akH) * np.log(
            (Q + 1 - kHsq + 2 * akH) / (Q + 1 - kHsq - 2 * akH + eps))
        teller = 256 * E * (Q + kHsq / 3 + 1 / 3) * np.exp(y)
        noemer = Zs ** 4 * pc.R() ** 2 * (
                    (Q - kHsq + 1) ** 2 + (4 * kHsq)) ** 3 + eps

        df_dE = teller / noemer
    return df_dE


def gdos_l(E, qa0sq, Z):
    """
     Calculates the hydrogenic GDOS for the l edge of an atom with atomic
     number Z.

     Parameters
     ----------
     E : float
         The energy in eV.
     qa0sq : float
         The q*a0**2 value (derived from momentum)
     Z : int
         Atomic number
     Returns
     -------
         GDOS
     """
    if Z == 6:
        # small test for the L edge of oxygen
        EL3 = 5
        EL1 = 7
        U = 0.5
    else:
        index = int(Z) - 13
        EL3 = IE3[index]
        EL1 = IE1[index]
        U = XU[index]

    Zs = Z - 0.35 * 7 - 1.7
    Q = (qa0sq / Zs ** 2)
    kHsq = E / (Zs ** 2 * pc.R()) - 0.25
    akH = np.sqrt(np.abs(kHsq))

    RF = ((E + 0.1 - EL3) / (1.8 * Z ** 2)) ** U

    if akH < 0.01:
        akH = 0.01

    if kHsq >= 0.0:
        beta = np.arctan(akH / (Q - kHsq + 0.25))
        if beta < 0:
            beta = beta + np.pi
        C = np.exp(-2 * beta / akH)
        D = 1 - np.exp(-2 * np.pi / akH)
    else:
        y = (-1 / akH) * np.log(
            (Q + 0.25 - kHsq + akH) / (Q + 0.25 - kHsq - akH))
        C = np.exp(y)
        D = 1.0
    if E <= EL1:
        G = 2.25 * Q ** 4 - (0.75 + 3.0 * kHsq) * Q ** 3 \
            + (0.59375 - 0.75 * kHsq - 0.5 * kHsq ** 2) * Q ** 2 \
            + (0.11146 + 0.85417 * kHsq + 1.8833 * kHsq ** 2 + kHsq ** 3) * Q \
            + 0.0035807 + kHsq / 21.333 + kHsq ** 2 / 4.5714 \
            + kHsq ** 3 / 2.4 + kHsq ** 4 / 4

        A = ((Q - kHsq + 0.25) ** 2 + kHsq) ** 5
    else:
        G = Q ** 3 - ((5 / 3) * kHsq + 11. / 12.) * Q ** 2 + \
            (kHsq ** 2 / 3 + 1.5 * kHsq + 65. / 48.) * Q + kHsq ** 3 / 3 \
            + 0.75 * kHsq ** 2 + kHsq * 23. / 48. + 5. / 64.
        A = ((Q - kHsq + 0.25) ** 2 + kHsq) ** 4

    df_dE = RF * 32 * G * C * E / (A * D * pc.R() ** 2 * Zs ** 4)

    return df_dE


def correction_factor_kohl(alpha, beta, theta, min_alpha=1e-6):
    """
    STILL NEEDS TO BE VALIDATED
    Calculates the correction factor when using a convergent
    probe. For probes having is convergence angle smaller than
    min_alpha no correction is applied.
    Ultramicroscopy 16 (1985) 265-268:
    https://doi.org/10.1016/0304-3991(85)90081-6
     Parameters
     ----------
     alpha : float
         Convergence angle in radians
     beta : float
         Collection angle in radians
     theta : float
         The angle for which the correction factor should be calculated
     min_alpha : float
        Minimum convergence angle for which the correction is applied

     Returns
     -------
     corr_factor : float
        correction factor used in the integration
    """
    thetasq = theta ** 2
    alphasq = alpha ** 2
    betasq = beta ** 2

    if alpha < min_alpha:
        corr_factor = 1.
    elif theta <= np.abs(alpha - beta):
        min_thetasq = min(alphasq, betasq)
        corr_factor = min_thetasq / alphasq
    # elif (theta>np.abs(alpha-beta))&(theta<alpha+beta):
    else:
        x = (alphasq + thetasq - betasq) / (2. * alpha * theta)
        y = (betasq + thetasq - alphasq) / (2. * beta * theta)
        wortel = np.sqrt(
            4 * alphasq * betasq - (alphasq + betasq - thetasq) ** 2)
        corr_factor = (1 / np.pi) * (
                    np.arccos(x) + (betasq / alphasq * np.arccos(y)) - (
                        1 / (2 * alphasq) * wortel))
    # else:
    #     corr_factor = 1

    return corr_factor


def convergence_correction_factor(alpha, beta, theta_sampling):
    """
    geometric correction factor for convergent beam in STEM
    #BY ZEZHONG
    Reference:

    Kohl, H. "A simple procedure for evaluating effective scattering
    cross-sections in STEM." Ultramicroscopy 16.2 (1985): 265-268.

    Args:
        alpha (float): incident beam convergence angle in rad
        beta (float): collection angle in rad
        theta_sampling (np.array): scattering angle in rad



    Returns:
        np.array: correction factor for each scattering angle
    """

    F = np.zeros(theta_sampling.shape)

    mask_1 = theta_sampling <= np.abs(alpha - beta)
    # print(mask_1)
    mask_2 = np.logical_and(theta_sampling > np.abs(alpha - beta),
                            theta_sampling < alpha + beta)
    # print(mask_2)
    mask_3 = theta_sampling >= (alpha + beta)
    # print(mask_3)

    x = (alpha ** 2 + beta ** 2 - theta_sampling[mask_2] ** 2) / (
                2 * alpha * beta)

    y = (beta ** 2 + theta_sampling[mask_2] ** 2 - alpha ** 2) / (
                2 * beta * theta_sampling[mask_2])

    F[mask_1] = min(alpha, beta) ** 2 / alpha ** 2

    term1 = np.arccos(x)
    term2 = (beta ** 2 / alpha ** 2) * np.arccos(y)
    wortel = np.sqrt(4 * alpha ** 2 * beta ** 2 - (
                (alpha ** 2 + beta ** 2 - theta_sampling[mask_2] ** 2) ** 2))
    term3 = (1 / 2 / alpha ** 2) * wortel

    F[mask_2] = (1 / np.pi) * (term1 + term2 - term3)

    F[mask_3] = 0

    return F


def dsigma_dE_hydrogenic(energy_axis, Z, ek, E0, beta, alpha, shell='K',
                         q_steps=100):
    """
    Calculates the differential cross section for a given energy axis. It uses
    the approximation of a convergent
    beam (alpha).


     Parameters
     ----------
     energy_axis : numpy array
         The energy in eV
     Z : int
         Atomic number
    ek: float
        The energy of the edge onset
    E0: float
        Energy of the incoming electron in eV
    beta: float
        Collection angle in radians
    alpha: float
        Convergence angle in radians (NOT YET IMPLEMENTED)
    shell: string
        Should be "K" or "L" to indicate which edge to calculate
    q_steps: int
        The number of calculations for every energy which is used to integrate
        over in q-space



     Returns
     -------
         dsigma_dE: numpy_array
            The differential energy cross section
    """

    if shell == 'K':
        gdos = gdos_k
    elif shell == 'L':
        gdos = gdos_l

    T = pc.T(E0)
    R = pc.R()
    gamma = pc.gamma(E0)
    dsigma_dE = np.zeros(energy_axis.size)
    for i in range(energy_axis.size):
        E = energy_axis[i]
        if E > ek:
            qa0sq_min, qa0sq_max = get_qmin_max(E, E0, beta, alpha=alpha)

            logqa0sq_axis = np.linspace(np.log(qa0sq_min), np.log(qa0sq_max),
                                        q_steps)
            lnqa0sqstep = (logqa0sq_axis[1] - logqa0sq_axis[0])
            integral = 0
            for j in range(logqa0sq_axis.size):
                qa0sq = np.exp(logqa0sq_axis[j])
                theta = 2. * np.sqrt(
                    np.abs(R * (qa0sq - qa0sq_min) / (4. * gamma ** 2 * T)))
                # integral += gdos(E, qa0sq, Z)*lnqa0sqstep
                integral += correction_factor_kohl(alpha, beta, theta) * gdos(
                    E, qa0sq, Z) * lnqa0sqstep  # correction does not work

            # dsigma_dE[i] = 4*np.pi*pc.a0()**2*(R/E)*(R/T)*integral*dispersion
            dsigma_dE[i] = 4 * np.pi * pc.a0() ** 2 * (R / E) * (
                        R / T) * integral

        else:
            dsigma_dE[i] = 0

    return dsigma_dE


def get_qmin_max(E, E0, beta, alpha=1e-9, return_q=False):
    """
    Returns the minimum and maximum q vector over which to integrate to
    get the cross section.

    """
    qa0sq_min = E ** 2 / (4 * pc.R() * pc.T(E0)) + (E ** 3) / (
                8 * pc.gamma(E0) ** 3 * pc.R() * pc.T(E0) ** 2)
    qa0sq_max = qa0sq_min + 4 * pc.gamma(E0) ** 2 * (pc.T(E0) / pc.R()) * (
        np.sin((beta + alpha) / 2)) ** 2
    if return_q:
        q_min = np.sqrt(qa0sq_min / pc.a0() ** 2)
        q_max = np.sqrt(qa0sq_max / pc.a0() ** 2)
        return q_min, q_max
    else:
        return qa0sq_min, qa0sq_max



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np


class Gaussian(Component):
    """
    A Gaussian component.

    Parameters
    ----------
    specshape : Spectrumshape
        The spectrum shape used to model

    A : float
        Amplitude of the gaussian function.

    centre: float
        The peak position of the gaussian function.

    fwhm: float
        The full width halve maximum of the gaussian function.

    Returns
    -------
    """

    def __init__(self, specshape, A, centre, fwhm):
        super().__init__(specshape)

        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(0, np.inf)
        p1.sethasgradient(True)
        self._addparameter(p1)

        p2 = Parameter('centre', centre)
        p2.setlinear(False)
        p2.sethasgradient(True)
        self._addparameter(p2)

        p3 = Parameter('fwhm', fwhm)
        p3.setlinear(False)
        p3.sethasgradient(True)
        p3.setboundaries(self.dispersion / 10,
                         np.inf)  # fwhm << dispersion makes no sense
        self._addparameter(p3)

        self._setname('Gaussian')

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p3 = self.parameters[2]
        if p1.ischanged() or p2.ischanged() or p3.ischanged():
            A = p1.getvalue()
            centre = p2.getvalue()
            fwhm = p3.getvalue()
            sigma = np.abs(fwhm) / (2.0 * np.sqrt(2.0 * np.log(2.0)))
            self.data = self.gaussian_function(A, centre, sigma)
        self.setunchanged()  # put parameters to unchanged

    def gaussian_function(self, A, centre, sigma):
        # if sigma<self.dispersion: #sigma < dispersion makes no sense
        #    sigma=self.dispersion
        return A * np.exp(-0.5 * ((self.energy_axis - centre) / sigma) ** 2)

    def getgradient(self, parameter):
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p3 = self.parameters[2]
        A = p1.getvalue()
        centre = p2.getvalue()
        fwhm = p3.getvalue()
        sigma = np.abs(fwhm) / (2.0 * np.sqrt(2.0 * np.log(2.0)))
        en = self.energy_axis

        if parameter == p1:
            self.gradient[0] = self.gaussian_function(1, centre, sigma)
            return self.gradient[0]
        if parameter == p2:
            self.gradient[1] = ((en - centre) / sigma ** 2) \
                               * self.gaussian_function(A, centre, sigma)
            return self.gradient[1]
        if parameter == p3:
            self.gradient[2] = (en - centre) ** 2 \
                / (sigma ** 2 * np.abs(fwhm)) \
                * self.gaussian_function(A, centre, sigma)
            return self.gradient[2]

        return None



from pyEELSMODEL.misc import hs_gdos as hsdos
from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge
from pyEELSMODEL.database.Segger_Guzzinati_Kohl.download import download_file
import os
import h5py
from pyEELSMODEL import __file__


class KohlLossEdge(CoreLossEdge):
    """
    Coreloss edges which are calculated by Leonhard Segger, Giulio Guzzinati
    and Helmut Kohl https://zenodo.org/record/6599071#.Y3I1cnbMKUk

    """

    def __init__(self, specshape, A, E0, alpha, beta, element, edge, eshift=0,
                 q_steps=100, dir_path=None,fast=False):
        """

        Parameters
        ----------
        specshape : Spectrumshape
        The spectrum shape used to model
        A : float
        Amplitude of the edge

        E0: float [V]
        The acceleration voltage of the microscope

        alpha: float [rad]
        The convergence angle of the incoming probe

        beta: float [rad]
        The collection angle

        element: string
        The name of the element from which to calculate the edge model.

        edge: string
        The type of edge. (K1, L3, M4, ...)

        eshift: float [eV]
        The shift of the onset energy with respect to the literature value.
        (default: 0)

        q_steps: int
        The number of q points taken into account for the integration over
        the momentum space. The larger the number of q_steps the more
        accurate the calculation. (default: 100)

        dir_path: string
            The filepath indicating where the GOS tables can be found.
            If None, the default path is used.

        fast: bool
        Use vectorized operations for the crossection calculation.

        Returns
        -------
        """
        if dir_path is None:
            dir_ = "/../pyEELSMODEL/database/Segger_Guzzinati_Kohl/"
            dir_path = os.path.dirname(
                os.path.dirname(__file__) + dir_
            )
        else:
            self.set_dir_path(dir_path)
        self.file = os.path.join(dir_path, 'Segger_Guzzinati_Kohl_1.5.0.gosh')
        if not os.path.exists(self.file):
            download_file(filename=self.file)
            

        self.set_dir_path(dir_path)
        super().__init__(specshape, A, E0, alpha, beta, element, edge,
                         eshift=eshift, q_steps=q_steps)

        self.set_gos_energy_q()
        self._fast = fast

    def set_gos_energy_q(self):
        if int(self.edge[-1]) == 1:
            edge = self.edge
        elif (int(self.edge[-1]) % 2 == 0):
            edge = self.edge + str(int(self.edge[-1]) + 1)
        else:
            edge = self.edge[0] + str(int(self.edge[-1]) - 1) + self.edge[-1]

        with h5py.File(self.file, 'r') as f:
            self.gos = f[self.element][edge]['data'][:]
            self.free_energies = f[self.element][edge]['free_energies'][:]
            self.q_axis = f[self.element][edge]['q'][:]

    def set_element(self, element):
        with h5py.File(self.file, 'r+') as f:
            elem_list = list(f.keys())
        if element in elem_list:
            self.element = element
        else:
            raise ValueError('Element you selected is not valid')

    def set_dir_path(self, path):
        self.dir_path = path
        self.file = os.path.join(self.dir_path,
                                 'Segger_Guzzinati_Kohl_1.5.0.gosh')

    # def set_onset_path(self, path):
    #     self.onset_path = path

    # def set_onset_energy(self):
    #     with h5py.File(self.onset_path, 'r+') as f:
    #         data = f[self.element][self.edge][:]
    #         self.onset_energy = data[0] + self.eshift
    #         self.prefactor = data[1]

    def _check_str_in_list(self, list, edge):
        for name in list:
            if name[0] == edge[0]:
                return True

        return False

    def set_edge(self, edge):
        """
        Checks if the given edge is valid and adds the directories of the
        :param edge:
        :return:
        """
        edge_list = ['K1', 'L1', 'L2', 'L3', 'M1', 'M2', 'M3', 'M4', 'M5',
                     'N1', 'N2', 'N3', 'N4', 'N5', 'N6', 'N7']
        if not isinstance(edge, str):
            raise TypeError('Edge should be a string: K1, L1, L2, L3, M2, M3,'
                            ' M4, M5, N4, N5', 'N6', 'N7')
        if edge in edge_list:
            self.edge = edge
        else:
            raise ValueError('Edge should be: K1, L1, L2, L3, M2, M3, M4, M5,'
                             ' N4, N5', 'N6', 'N7')

    # def set_edge(self, edge):
    #     """
    #     Checks if the given edge is valid and adds the directories of the
    #     :param edge:
    #     :return:
    #     """
    #     edge_list = ['K1', 'L1', 'L2', 'L3', 'M1', 'M2', 'M3', 'M4', 'M5',
    #                  'N1', 'N2', 'N3', 'N4', 'N5', 'N6', 'N7']
    #
    #     if not isinstance(edge, str):
    #         raise TypeError('Edge should be a string: K1, L1, L2, L3, M2,'
    #                         ' M3,'
    #                         ' M4, M5, N4, N5', 'N6', 'N7')
    #     if edge in edge_list:
    #         self.edge = edge
    #     else:
    #         raise ValueError('Edge should be: K1, L1, L2, L3, M2, M3, M4,'
    #                          ' M5,'
    #                          ' N4, N5', 'N6', 'N7')

    def calculate_cross_section(self):
        """
        Calculates the cross section in barns (1e-28 m^2)


        """
        ek = self.onset_energy
        E0 = self.parameters[1].getvalue()
        alpha = self.parameters[3].getvalue()
        beta = self.parameters[2].getvalue()
        e_axis = self.free_energies
        q_axis = self.q_axis
        gos = self.gos

        prf = 1e28 * self.prefactor  # prefactor and convert to barns

        if self._fast:
            css = prf * hsdos.dsigma_dE_from_GOSarray_FastKohl(self.energy_axis,
                                                  e_axis + ek, ek, E0, beta,
                                                  alpha, q_axis, gos,
                                                  q_steps=100)

        else:
            css = prf * hsdos.dsigma_dE_from_GOSarray(self.energy_axis,
                                                      e_axis + ek, ek, E0, beta,
                                                      alpha, q_axis, gos,
                                                      q_steps=100,
                                                      swap_axes=True)

        return css
